@@ -1,6 +1,5 @@
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE NoImplicitPrelude #-}
-{-# OPTIONS_GHC -fno-warn-dodgy-exports #-} -- Char8
 
 -- |
 -- Module      : Network.TLS.Imports
@@ -13,7 +12,7 @@ module Network.TLS.Imports
     (
     -- generic exports
       ByteString
-    , module Data.ByteString.Char8 -- instance
+    , (<&>)
     , module Control.Applicative
     , module Control.Monad
 #if !MIN_VERSION_base(4,13,0)
@@ -22,22 +21,18 @@ module Network.TLS.Imports
     , module Data.Bits
     , module Data.List
     , module Data.Maybe
-#if MIN_VERSION_base(4,9,0)
     , module Data.Semigroup
-#else
-    , module Data.Monoid
-#endif
     , module Data.Ord
     , module Data.Word
-#if !MIN_VERSION_base(4,8,0)
-    , sortOn
-#endif
     -- project definition
     , showBytesHex
     ) where
 
 import Data.ByteString (ByteString)
-import Data.ByteString.Char8 ()
+import Data.ByteString.Char8 () -- instance
+#if MIN_VERSION_base(4,11,0)
+import Data.Functor
+#endif
 
 import Control.Applicative
 import Control.Monad
@@ -47,23 +42,17 @@ import Control.Monad.Fail (MonadFail)
 import Data.Bits
 import Data.List
 import Data.Maybe hiding (fromJust)
-#if MIN_VERSION_base(4,9,0)
 import Data.Semigroup
-#else
-import Data.Monoid
-#endif
 import Data.Ord
 import Data.Word
 
 import Data.ByteArray.Encoding as B
 import qualified Prelude as P
 
-#if !MIN_VERSION_base(4,8,0)
-import Prelude ((.))
-
-sortOn :: Ord b => (a -> b) -> [a] -> [a]
-sortOn f =
-  map P.snd . sortBy (comparing P.fst) . map (\x -> let y = f x in y `P.seq` (y, x))
+#if !MIN_VERSION_base(4,11,0)
+(<&>) :: Functor f => f a -> (a -> b) -> f b
+(<&>) = P.flip fmap
+infixl 1 <&>
 #endif
 
 showBytesHex :: ByteString -> P.String
